/* $LAAS: correlation_census.c 2010/12/10 11:41:49 mallet $ */

/*
 * Copyright (c) 2008,2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *					Anthony Mallet on Thu Sep  4 2008
 */

static const char _rcsid[] =
  "$LAAS: correlation_census.c 2010/12/10 11:41:49 mallet $";

#define SPIX_CORREL_NAME		census
#define SPIX_CORREL_COMBINE_TYPE	short
#define SPIX_CORREL_SUM_TYPE		short
#define SPIX_CORREL_SCORE_TYPE		short
#define SPIX_CORREL_COMBINE_VEC		1

#include "correlation.tmpl.c"

/* pre-compute count of unset bits for integers from 0 to 255 */
#define gen_unsetbitcount_1(n)	n+1,			  n
#define gen_unsetbitcount_2(n)	gen_unsetbitcount_1(n+1), gen_unsetbitcount_1(n)
#define gen_unsetbitcount_3(n)	gen_unsetbitcount_2(n+1), gen_unsetbitcount_2(n)
#define gen_unsetbitcount_4(n)	gen_unsetbitcount_3(n+1), gen_unsetbitcount_3(n)
#define gen_unsetbitcount_5(n)	gen_unsetbitcount_4(n+1), gen_unsetbitcount_4(n)
#define gen_unsetbitcount_6(n)	gen_unsetbitcount_5(n+1), gen_unsetbitcount_5(n)
#define gen_unsetbitcount_7(n)	gen_unsetbitcount_6(n+1), gen_unsetbitcount_6(n)
#define gen_unsetbitcount_8	gen_unsetbitcount_7(1),	  gen_unsetbitcount_7(0)

static const SPIX_CORREL_COMBINE_TYPE unsetbitcount[1<<8] = { gen_unsetbitcount_8 };


static inline int
spix_correl_chkparam(
  unsigned char *source_gauche, unsigned char *source_droite,
  SP_IMAGES_TMP *images_gauches, SP_IMAGES_TMP *images_droites,
  SP_CORRELATION_PARAMS *correlation_params,
  SP_CORRELATION_RESULTS *correlation_results)
{
  if ((images_gauches->state_moyennes != SP_IMAGE_ALLOCATED)	||
      (images_droites->state_moyennes != SP_IMAGE_ALLOCATED))
    return stereoError(ERR_SP_IMAGE_TMP_INVALIDE);

  return STEREO_OK;
}


static int
spix_preproc_census(unsigned char *source, SP_IMAGES_TMP *imtmp)
{
  int maskpos[8];
  int offset;
  int i, j, k, c;

  unsigned char *maskval[8];
  unsigned char *pivot_ptr;
  unsigned char pivot;
  unsigned char *im_census_ptr;

  /* Offset from current position for 3x3 mask */
  /* going clockwise starting top-left */
  maskpos[0] = - imtmp->images_sizes.nb_col_final - 1;
  maskpos[1] = - imtmp->images_sizes.nb_col_final * 2;
  maskpos[2] = - imtmp->images_sizes.nb_col_final + 1;
  maskpos[3] =   2;
  maskpos[4] =   imtmp->images_sizes.nb_col_final + 1;
  maskpos[5] =   imtmp->images_sizes.nb_col_final * 2;
  maskpos[6] =   imtmp->images_sizes.nb_col_final - 1;
  maskpos[7] = - 2;

  for(i = imtmp->images_sizes.i_min + 2; i < imtmp->images_sizes.i_max - 2; i++) {
    offset = i * imtmp->images_sizes.nb_col_final + imtmp->images_sizes.j_min + 2;

    im_census_ptr = ((unsigned char *)imtmp->intermediaire_moyennes1) + offset;
    pivot_ptr = source + offset;
    for(k=0; k<8; k++) maskval[k] = pivot_ptr + maskpos[k];

    for (j = imtmp->images_sizes.j_min + 2; j < imtmp->images_sizes.j_max - 2; j++) {
      pivot = *(pivot_ptr++);

      c = 0;
      for(k=0; k<8; k++) if (*(maskval[k]++) < pivot) c += (1<<k);
      *(im_census_ptr++) = (unsigned char)c;
    }
  }

  return STEREO_OK;
}

static inline int
spix_correl_preproc(
  unsigned char **source_gauche, unsigned char **source_droite,
  SP_IMAGES_TMP *images_gauches, SP_IMAGES_TMP *images_droites)
{
  spix_preproc_census(*source_gauche, images_gauches);
  spix_preproc_census(*source_droite, images_droites);

  *source_gauche = (unsigned char *)images_gauches->intermediaire_moyennes1;
  *source_droite = (unsigned char *)images_droites->intermediaire_moyennes1;

  return STEREO_OK;
}

static inline SPIX_CORREL_COMBINE_TYPE
spix_correl_pixel_combine(unsigned char a, unsigned char b)
{
  return unsetbitcount[a ^ b];
}

static inline void
spix_correl_pixel_combine_vec(SPIX_CORREL_COMBINE_TYPE *c,
			      spix_vector(unsigned char) a,
			      spix_vector(unsigned char) b)
{
#if SPIX_SSE && defined(__SSE2__) && defined(__GNUC__)
  __m128i k0 = _mm_set1_epi8(0);
  __m128i k1 = _mm_set1_epi8(0x55);
  __m128i k2 = _mm_set1_epi8(0x33);
  __m128i k4 = _mm_set1_epi8(0x0f);

  /* compute bits that differ in a and b */
  __m128i x = ~(a ^ b);

  /* count bits in x bytewise */
  x =  x       - (_mm_srli_epi16(x,1) & k1);	/* count duo bits with x - x/2 */
  x = (x & k2) + (_mm_srli_epi16(x,2) & k2);	/* sum duo bits into nibbles */
  x = (x       + (_mm_srli_epi16(x,4))) & k4;	/* sum nibbles into bytes */

  spix_stru(c, SPIX_CORREL_COMBINE_TYPE, _mm_unpacklo_epi8(x, k0));
  spix_stru(c + spix_vector_len(SPIX_CORREL_COMBINE_TYPE),
	    SPIX_CORREL_COMBINE_TYPE, _mm_unpackhi_epi8(x, k0));
#else
  *c = unsetbitcount[a ^ b];
#endif
}

static inline SPIX_CORREL_SCORE_TYPE
spix_correl_score(SPIX_CORREL_SUM_TYPE sum, double norm,
		  float lavg, float ravg, float lrstddev, float rrstddev)
{
  return sum;
}

static inline spix_vector(SPIX_CORREL_SCORE_TYPE)
spix_correl_score_vec(spix_vector(SPIX_CORREL_SUM_TYPE) sum, double norm,
		      spix_vector(float) lavg, spix_vector(float) ravg,
		      spix_vector(float) lrstddev, spix_vector(float) rrstddev)
{
  return sum;
}
