/* $LAAS: correlation_zncc.c 2010/12/10 11:44:29 mallet $ */

/* 
 * Copyright (c) 2008,2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *					Anthony Mallet on Thu Sep  4 2008
 */

static const char _rcsid[] =
  "$LAAS: correlation_zncc.c 2010/12/10 11:44:29 mallet $";

#define SPIX_CORREL_NAME		zncc
#define SPIX_CORREL_COMBINE_TYPE	int
#define SPIX_CORREL_SUM_TYPE		int
#define SPIX_CORREL_SCORE_TYPE		float

#include "correlation.tmpl.c"

static inline int
spix_correl_chkparam(
  unsigned char *source_gauche, unsigned char *source_droite,
  SP_IMAGES_TMP *images_gauches, SP_IMAGES_TMP *images_droites,
  SP_CORRELATION_PARAMS *correlation_params,
  SP_CORRELATION_RESULTS *correlation_results)
{
  if ((images_gauches->state_moyennes != SP_IMAGE_COMPUTED)	||
      (images_droites->state_moyennes != SP_IMAGE_COMPUTED)	||
      (images_gauches->state_sigmas != SP_IMAGE_COMPUTED)	||
      (images_droites->state_sigmas != SP_IMAGE_COMPUTED))
    return stereoError(ERR_SP_IMAGE_TMP_INVALIDE);

  return STEREO_OK;
}

static inline int
spix_correl_preproc(
  unsigned char **source_gauche, unsigned char **source_droite,
  SP_IMAGES_TMP *images_gauches, SP_IMAGES_TMP *images_droites)
{
  return STEREO_OK;
}

static inline SPIX_CORREL_COMBINE_TYPE
spix_correl_pixel_combine(unsigned char a, unsigned char b)
{
  return a * b;
}

static inline void
spix_correl_pixel_combine_vec(SPIX_CORREL_COMBINE_TYPE *c,
			      spix_vector(unsigned char) a,
			      spix_vector(unsigned char) b)
{
#if SPIX_SSE && defined(__SSE2__) && defined(__GNUC__)
  __m128i k0 = _mm_set1_epi8(0);
  __m128i al = _mm_unpacklo_epi8(a, k0);
  __m128i ah = _mm_unpackhi_epi8(a, k0);
  __m128i bl = _mm_unpacklo_epi8(b, k0);
  __m128i bh = _mm_unpackhi_epi8(b, k0);

  /* multiply a by b */
  __m128i xl = _mm_mullo_epi16(al, bl);
  __m128i xh = _mm_mullo_epi16(ah, bh);

  spix_stru(c, SPIX_CORREL_SUM_TYPE,
	    _mm_unpacklo_epi16(xl, k0));
  spix_stru(c + spix_vector_len(SPIX_CORREL_SUM_TYPE), SPIX_CORREL_SUM_TYPE,
	    _mm_unpackhi_epi16(xl, k0));
  spix_stru(c + 2*spix_vector_len(SPIX_CORREL_SUM_TYPE), SPIX_CORREL_SUM_TYPE,
	    _mm_unpacklo_epi16(xh, k0));
  spix_stru(c + 3*spix_vector_len(SPIX_CORREL_SUM_TYPE), SPIX_CORREL_SUM_TYPE,
	    _mm_unpackhi_epi16(xh, k0));
#else
  *c = a * b;
#endif
}

static inline SPIX_CORREL_SCORE_TYPE
spix_correl_score(SPIX_CORREL_SUM_TYPE sum, double norm,
		  float lavg, float ravg, float lrstddev, float rrstddev)
{
  return (sum * norm - lavg * ravg) * lrstddev * rrstddev;
}

static inline spix_vector(SPIX_CORREL_SCORE_TYPE)
spix_correl_score_vec(spix_vector(SPIX_CORREL_SUM_TYPE) sum, double norm,
		      spix_vector(float) lavg, spix_vector(float) ravg,
		      spix_vector(float) lrstddev, spix_vector(float) rrstddev)
{
  return (spix_cvtips(&sum) * spix_1float(norm) - lavg * ravg) * lrstddev * rrstddev;
}
