/*	$LAAS$ */

/*
 * Copyright (c) 2000,2008,2010 LAAS/CNRS                       --  Fri Aug  4 2006
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#include "stereopixel-config.h"

/* librairies standards */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Les erreurs, structures et numrations locales */
#include "libstereopixel.h"

#define MAX(a,b)	((a)>(b)?(a):(b))
#define MIN(a,b)	((a)<(b)?(a):(b))

/********************************************************************
 *
 * Cette fonction dtermine les tableaux de disparit min et max pour chaque
 * colonne de l'image.
 *
 ********************************************************************/

void
calcule_disp_tableaux (int     nbcol,
		       int     nj2,
		       int     disparite_min,
		       int     disparite_max,
		       int*    tab_disp_max,
		       int*    tab_disp_min,
		       int*    tab_disp_nj2_max,
		       int*    tab_disp_nj2_min,
		       int*    disp_max_general,
		       int*    disp_min_general)

{
  int    j;			/* Pour boucler sur les colonnes */
  int*   disp_min_ptr;
  int*   disp_max_ptr;
  int*   disp_min_nj2_ptr;
  int*   disp_max_nj2_ptr;	/* Pointeurs sur les tableaux */


  /* On initialise les valeurs min et max gnrales */
  *disp_min_general = nbcol;
  *disp_max_general = - nbcol;

  /* On positionne les pointeurs au dbut des tableaux */
  disp_max_ptr = tab_disp_max;
  disp_min_ptr = tab_disp_min;
  disp_max_nj2_ptr = tab_disp_nj2_max;
  disp_min_nj2_ptr = tab_disp_nj2_min;

  /* On boucle sur les lments de la ligne */
  for (j = 0; j < nbcol; j++) {

    *disp_min_ptr = MAX (disparite_min, - j);
    *disp_max_ptr = MIN (disparite_max, nbcol - j);

    *disp_min_nj2_ptr = MAX (disparite_min, nj2 - j);
    *disp_max_nj2_ptr = MIN (disparite_max, nbcol - j - nj2);

    *disp_max_general = MAX(*disp_max_ptr, *disp_max_general);
    *disp_min_general = MIN(*disp_min_ptr, *disp_min_general);

    disp_max_ptr++;
    disp_min_ptr++;
    disp_max_nj2_ptr++;
    disp_min_nj2_ptr++;
  }

#if SPIX_SSE && defined(__SSE2__) && defined(__GNUC__)
  /* in order to align all the arrays pointer and synchronize the
   * different arrays, we need delta_max_disp*sizeof(*ptr) to be a
   * multiple of 16 bytes. We'd also like 'disp_min_nj2 + 1 - disp_min_general'
   * to be a multiple of 16 bytes for each pixel so that the correlation
   * loop doesn't have to deal with unaligned data. Due to the optimizations in
   * the score computation, it's not possible to satisfy this constraint.
   * So we just align disp_min_general so that most of the correlations will
   * start on a 16 bytes boundaries. The two lines below save ~10% of execution
   * time.
   *
   * To be on the safe side we make those value a multiple of 8 so that
   * arrays of shorts, ints and floats are all ok. */

  /* adjust disp_min_general */
  *disp_min_general -= 15;

  /* round disp_max_general - disp_min_general */
  if ((*disp_max_general - *disp_min_general + 1) & 15)
    *disp_max_general = 15 +
      ((*disp_max_general - *disp_min_general + 1) & ~15) + *disp_min_general;
#endif
}


