/*	$LAAS: correlation_results.c 2011/01/06 11:38:27 mallet $ */

/*
 * Copyright (c) 2000,2008,2010-2011 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                                        Fri Aug  4 2006
 */

#include "stereopixel-config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "libstereopixel.h"
#include "vectorization.h"
#if SPIX_THREADS
# include "workers.h"
#endif

/* Provide a replacement for posix_memalign on macosx < 10.6 */
#ifndef HAVE_POSIX_MEMALIGN
# ifdef __APPLE__
/* malloc in Mac OS X guarantees 16 byte alignment. */
int
posix_memalign(void **memptr, size_t alignment, size_t size)
{
  *memptr = malloc(size);
  return (*memptr == NULL)?ENOMEM:0;
}
# else
#  error Need posix_memalign implementation
# endif /* __APPLE__ */
#endif /* HAVE_POSIX_MEMALIGN */


/***********************************************************************
 *
 * Fonction de cration d'une structure 'SP_CORRELATION_RESULTS'
 *
 ***********************************************************************/

SP_CORRELATION_RESULTS*
create_correlation_results () {
  SP_CORRELATION_RESULTS* new_correlation_results;

  new_correlation_results
    = (SP_CORRELATION_RESULTS*) malloc (sizeof (SP_CORRELATION_RESULTS));

  if (new_correlation_results == NULL) {
    stereoError (ERR_SP_MALLOC_CORRELATION_RESULTS);
    return (NULL);
  }

  new_correlation_results->state = SP_CORRELATION_RESULTS_NOT_ALLOCATED;

  return (new_correlation_results);
}

/***********************************************************************
 *
 * Fonction de destruction d'une structure 'SP_CORRELATION_RESULTS'
 *
 ***********************************************************************/

void
destroy_correlation_results (SP_CORRELATION_RESULTS* correlation_results)
{

  if (correlation_results == NULL)
    return;

  sc_clean_correlation_results(correlation_results);
  free (correlation_results);

  return;
}

void
sc_clean_correlation_results(SP_CORRELATION_RESULTS *correlation_results)
{
   if (correlation_results != NULL &&
       correlation_results->state == SP_CORRELATION_RESULTS_ALLOCATED) {
      int i;

      free (correlation_results->tab_disp_min);
      free (correlation_results->tab_disp_max);
      free (correlation_results->tab_disp_nj2_min);
      free (correlation_results->tab_disp_nj2_max);

      for (i = 0; i < correlation_results->win_nb_lig+1; i++)
	 free (* (correlation_results->tab_tab_produits + i));
      free (correlation_results->tab_tab_produits);

      free (correlation_results->tab_scores);
      free (correlation_results->tab_scores_2);

      free (correlation_results->tab_max);

      free (correlation_results->somme_pdt_bas);
      free (correlation_results->somme_pdt_haut);

      free (correlation_results->im_disparites);
      free (correlation_results->im_sigma_disparites);
      free (correlation_results->im_regions);

      correlation_results->state = SP_CORRELATION_RESULTS_NOT_ALLOCATED;
   }
}

/***********************************************************************
 *
 * Fonction de copie d'une structure 'SP_CORRELATION_RESULTS'
 *
 ***********************************************************************/

SP_CORRELATION_RESULTS*
copy_correlation_results (SP_CORRELATION_RESULTS* correlation_results)
{
/*   SP_CORRELATION_RESULTS* new_correlation_results; */

  if (correlation_results == NULL)
    return (NULL);

  printf ("ERR_SP_FCT_NON_IMPLEMENTEE");
  return (NULL);

/*   new_correlation_results = create_correlation_results (); */

/*   memcpy (new_correlation_results, correlation_results, sizeof (SP_CORRELATION_RESULTS)); */

/*   return (new_correlation_results); */
}

/***********************************************************************
 *
 * Affiche des infos sur une structure 'SP_CORRELATION_RESULTS'
 *
 ***********************************************************************/

int
info_correlation_results (SP_CORRELATION_RESULTS* correlation_results)
{
  switch (correlation_results->state) {
  case SP_CORRELATION_RESULTS_NOT_ALLOCATED:
    printf ("Structure 'SP_CORRELATION_RESULTS' pas alloue\n");
    break;
  case SP_CORRELATION_RESULTS_ALLOCATED:
    printf ("Structure 'SP_CORRELATION_RESULTS' alloue :\n");

    /* tat de l'image de disparits */
    switch (correlation_results->state_disparites) {
    case SP_IMAGE_ALLOCATED:
      printf ("\tImage de disparits simplement alloue\n");
      break;
    case SP_IMAGE_COMPUTED:
      printf ("\tImage de disparits dtermine\n");
      break;
    default:
      return stereoError(ERR_SP_CAS_IMPOSSIBLE);
      /*     break; */
      /* Unreached */
    }

    /* tat de l'image des carts types sur les disparits */
    switch (correlation_results->state_sigma_disparites) {
    case SP_IMAGE_ALLOCATED:
      printf ("\tImage des sigma-disparits simplement alloue\n");
      break;
    case SP_IMAGE_COMPUTED:
      printf ("\tImage des sigma-disparits dtermine\n");
      break;
    default:
      return stereoError(ERR_SP_CAS_IMPOSSIBLE);
      /*     break; */
      /* Unreached */
    }

    /* tats de l'image des rgions  */
    switch (correlation_results->state_regions) {
    case SP_IMAGE_ALLOCATED:
      printf ("\tImage des rgions simplement alloue\n");
      break;
    case SP_IMAGE_COMPUTED:
      printf ("\tImage des rgions dtermine\n");
      break;
    default:
      return stereoError(ERR_SP_CAS_IMPOSSIBLE);
      /*     break; */
      /* Unreached */
    }

    break;
  default:
    return stereoError(ERR_SP_CAS_IMPOSSIBLE);
    /*     break; */
    /* Unreached */
  }



  return (STEREO_OK);
}

/***********************************************************************
 *
 * Fonction d'initialisation d'une structure 'SP_CORRELATION_RESULTS'
 *
 ***********************************************************************/
int
init_correlation_results (SP_IMAGES_SIZES*		images_sizes,
			  SP_CORRELATION_PARAMS*	correlation_params,
			  SP_CORRELATION_RESULTS*	correlation_results)

{
  /* Variables extraites des paramtres */
  int    nblig, nbcol;
  int    taux_reduction_j;

  int    disparite_min;
  int    disparite_max;

  /* Variables dtermines */
  int    disp_max_general;
  int    disp_min_general;
  int    delta_max_disp;
  int    nj2;
  int    ni;
  int    score2_tab_size;	/* (delta_max_disp * nbcol * sizeof(double)) */
  int    score_tab_size;	/* (delta_max_disp * nbcol * sizeof(double)) */
  int    produits_tab_size;	/* (delta_max_disp * nbcol * sizeof(int)) */
  int    sommes_tab_size;	/* ( delta_max_disp * sizeof (int)) */

  /* Variables diverses et avaries */
  int    i, s;
  void *mem;

  /* Test de validit des paramtres */
  if (images_sizes->state != SP_IMAGES_SIZES_INITIALIZED)
    return stereoError(ERR_SP_IMAGES_SIZES_NOT_INITIALIZED);

  if (correlation_params->state != SP_CORRELATION_PARAMS_INITIALIZED)
    return stereoError(ERR_SP_CORRELATION_PARAMS_NOT_INITIALIZED);

  /* Copie de la structure 'images_sizes' dans le champ adquat de la
     structure images_results */
  memcpy (&(correlation_results->images_sizes), images_sizes, sizeof (SP_IMAGES_SIZES));

  /* On rcupre les variables des paramtres */
  nblig = images_sizes->nb_lig_final;
  nbcol = images_sizes->nb_col_final;
  taux_reduction_j = images_sizes->taux_reduction_j;

  disparite_min = correlation_params->disp_min;
  disparite_max = correlation_params->disp_max;
  ni = correlation_params->win_nb_lig;

  /* On alloue les tableaux de valeurs min et max de la disparit */
  correlation_results->tab_disp_min = (int*) malloc (nbcol * sizeof (int));
  if (correlation_results->tab_disp_min == NULL)
    return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
  correlation_results->tab_disp_max = (int*) malloc (nbcol * sizeof (int));
  if (correlation_results->tab_disp_max == NULL)
    return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
  correlation_results->tab_disp_nj2_min = (int*) malloc (nbcol * sizeof (int));
  if (correlation_results->tab_disp_nj2_min == NULL)
    return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
  correlation_results->tab_disp_nj2_max = (int*) malloc (nbcol * sizeof (int));
  if (correlation_results->tab_disp_nj2_max == NULL)
    return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);

  /* On dtermine les tableaux de disparit */
  nj2 = correlation_params->win_nb_col / 2;
  calcule_disp_tableaux (nbcol, nj2,
			 disparite_min / taux_reduction_j,
			 disparite_max / taux_reduction_j,
			 correlation_results->tab_disp_max,
			 correlation_results->tab_disp_min,
			 correlation_results->tab_disp_nj2_max,
			 correlation_results->tab_disp_nj2_min,
			 &disp_max_general, &disp_min_general);
  delta_max_disp = disp_max_general - disp_min_general + 1;

  correlation_results->delta_max_disp = delta_max_disp;
  correlation_results->disp_min_general = disp_min_general;
  correlation_results->win_nb_lig = ni;

  /* allocate memory for score arrays. We allocate extra bytes because
   * some vectorized computations might also produce extra data at the
   * end of lines that are not multiple of 128bits in length. */
  switch(correlation_params->type_correlation) {
    case SP_CORRELATION_ZNCC:
    case SP_CORRELATION_ZNCC_SIGMA:
      score2_tab_size = delta_max_disp * nbcol * sizeof(float) + spix_align_size;
      produits_tab_size = delta_max_disp * nbcol * sizeof(int) + 4*spix_align_size;
      score_tab_size = delta_max_disp * nbcol * sizeof(int) + spix_align_size;
      sommes_tab_size = delta_max_disp * sizeof(int) + spix_align_size;
      break;

    case SP_CORRELATION_CENSUS:
    case SP_CORRELATION_CENSUS_SIGMA:
      score2_tab_size = delta_max_disp * nbcol * sizeof(short) + spix_align_size;
      produits_tab_size = delta_max_disp * nbcol * sizeof(short) + spix_align_size;
      score_tab_size = delta_max_disp * nbcol * sizeof(short) + spix_align_size;
      sommes_tab_size = delta_max_disp * sizeof(short) + spix_align_size;
      break;

    default:
      score2_tab_size = 0;
      score_tab_size = 0;
      produits_tab_size = 0;
      sommes_tab_size = 0;
      break;
  }

  s = posix_memalign(&mem, spix_align_size, score2_tab_size);
  if (s) return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
  correlation_results->tab_scores_2 = mem;

  s = posix_memalign(&mem, spix_align_size, sommes_tab_size);
  if (s) return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
  correlation_results->somme_pdt_bas = mem;

  s = posix_memalign(&mem, spix_align_size, sommes_tab_size);
  if (s) return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
  correlation_results->somme_pdt_haut = mem;

  correlation_results->tab_tab_produits = malloc((ni+1) * sizeof(int *));
  if (correlation_results->tab_tab_produits == NULL)
    return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
  for (i = 0; i < ni+1; i++) {
    s = posix_memalign(&mem, spix_align_size, produits_tab_size);
    if (s) return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
    correlation_results->tab_tab_produits[i] = mem;
  }

  s = posix_memalign(&mem, spix_align_size, score_tab_size);
  if (s) return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
  correlation_results->tab_scores = mem;

  correlation_results->tab_max = malloc(nbcol * sizeof(int));
  if (!correlation_results->tab_max)
    return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);

#if SPIX_THREADS
  if (correlation_params->threads > 1) {
    spix_worker *w = NULL;
    int t, delta, k;

    delta = (nblig - ni) / correlation_params->threads;
    k = ni/2 + delta;

    for(t=0; t<correlation_params->threads-1; t++) {
      w = correlation_params->thread_array[t];

      w->i_min = k;
      w->i_max = k + delta;
      k += delta;

      s = posix_memalign(&mem, spix_align_size, score2_tab_size);
      if (s) return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
      w->tab_scores_2 = mem;

      s = posix_memalign(&mem, spix_align_size, sommes_tab_size);
      if (s) return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
      w->somme_pdt_bas = mem;

      s = posix_memalign(&mem, spix_align_size, sommes_tab_size);
      if (s) return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
      w->somme_pdt_haut = mem;

      w->tab_tab_produits = malloc((ni+1) * sizeof(int *));
      if (w->tab_tab_produits == NULL)
	return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
      for (i = 0; i < ni+1; i++) {
	s = posix_memalign(&mem, spix_align_size, produits_tab_size);
	if (s) return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
	w->tab_tab_produits[i] = mem;
      }

      s = posix_memalign(&mem, spix_align_size, score_tab_size);
      if (s) return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
      w->tab_scores = mem;

      w->tab_max = malloc(nbcol * sizeof(int));
      if (!w->tab_max)
	return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
    }
    w->i_max = nblig - ni/2;
  }
#endif

  /* Suivent toutes les allocations des images rsultats : l'image de disparits... */
  correlation_results->im_disparites
    = (float*) malloc (images_sizes->taille_finale * sizeof (float));
  if (correlation_results->im_disparites == NULL)
    return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
  correlation_results->state_disparites = SP_IMAGE_ALLOCATED;

  /* ... l'image des cats types sur les disparits... */
  correlation_results->im_sigma_disparites
    = (float*) malloc (images_sizes->taille_finale * sizeof (float));
  if (correlation_results->im_sigma_disparites == NULL)
    return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
  correlation_results->state_sigma_disparites = SP_IMAGE_ALLOCATED;

  /* ... et l'image des rgions. */
  correlation_results->im_regions
    = (int*) malloc (images_sizes->taille_finale * sizeof (int));
  if (correlation_results->im_regions == NULL)
    return stereoError(ERR_SP_MALLOC_CORRELATION_RESULTS);
  correlation_results->state_regions = SP_IMAGE_ALLOCATED;

  correlation_results->state = SP_CORRELATION_RESULTS_ALLOCATED;

  /* Set to 0 the correlation quality criteria */
  correlation_results->nbRemovedByUnicity = 0;
  correlation_results->nbRemovedByBlob = 0;
  correlation_results->nbCorrelated = 0;

  return (STEREO_OK);
}
