/*	$LAAS$ */

/*
 * Copyright (c) 2000 LAAS/CNRS                       --  Fri Aug  4 2006
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

/* librairies standards */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Les prototypes */
#include "libstereopixel.h"

/***********************************************************************
 *
 * Fonction de cration d'une structure 'SP_IMAGES_TMP'
 *
 ***********************************************************************/

SP_IMAGES_TMP*
create_images_tmp () {
  SP_IMAGES_TMP* new_images_tmp;

  new_images_tmp = (SP_IMAGES_TMP*) malloc (sizeof (SP_IMAGES_TMP));

  if (new_images_tmp == NULL) {
    stereoError (ERR_SP_MALLOC_IMAGES_TMP);
    return (NULL);
  }

  new_images_tmp->state = SP_IMAGES_TMP_NOT_ALLOCATED;

  return (new_images_tmp);
}

/***********************************************************************
 *
 * Fonction de destruction d'une structure 'SP_IMAGES_TMP'
 *
 ***********************************************************************/

void
destroy_images_tmp (SP_IMAGES_TMP* images_tmp)
{
  if (images_tmp == NULL)
    return;

  sc_clean_images_tmp(images_tmp);
  free (images_tmp);

  return;
}

void
sc_clean_images_tmp(SP_IMAGES_TMP *images_tmp)
{
  if (images_tmp != NULL && images_tmp->state == SP_IMAGES_TMP_ALLOCATED) {

    free (images_tmp->intermediaire_haut);
    free (images_tmp->intermediaire_bas);
    free (images_tmp->intermediaire_moyennes1);
    free (images_tmp->intermediaire_moyennes2);
    free (images_tmp->im_moyennes);
    free (images_tmp->im_sigmas);
  }
}

/***********************************************************************
 *
 * Fonction de copie d'une structure 'SP_IMAGES_TMP'
 *
 ***********************************************************************/

SP_IMAGES_TMP*
copy_images_tmp (SP_IMAGES_TMP* images_tmp)
{
  SP_IMAGES_TMP* new_images_tmp;

  if (images_tmp == NULL)
    return (NULL);

  new_images_tmp = create_images_tmp ();

  memcpy (new_images_tmp, images_tmp, sizeof (SP_IMAGES_TMP));

  if (images_tmp->state == SP_IMAGES_TMP_ALLOCATED) {
    /* Copie des images de la structure */
    memcpy (new_images_tmp->intermediaire_haut,
	    images_tmp->intermediaire_haut,
	    images_tmp->images_sizes.taille_finale * sizeof (short));

    memcpy (new_images_tmp->intermediaire_bas,
	    images_tmp->intermediaire_bas,
	    images_tmp->images_sizes.taille_finale * sizeof (short));

    memcpy (new_images_tmp->intermediaire_moyennes1,
	    images_tmp->intermediaire_moyennes1,
	    images_tmp->images_sizes.taille_finale * sizeof (short));

    memcpy (new_images_tmp->intermediaire_moyennes2,
	    images_tmp->intermediaire_moyennes2,
	    images_tmp->images_sizes.taille_finale * sizeof (int));

    memcpy (new_images_tmp->im_moyennes,
	    images_tmp->im_moyennes,
	    images_tmp->images_sizes.taille_finale * sizeof (float));

    memcpy (new_images_tmp->im_sigmas,
	    images_tmp->im_sigmas,
	    images_tmp->images_sizes.taille_finale * sizeof (float));
  }

  return (new_images_tmp);
}

/***********************************************************************
 *
 * Affiche des infos sur une structure 'SP_IMAGES_TMP'
 *
 ***********************************************************************/

int
info_images_tmp (SP_IMAGES_TMP* images_tmp)
{
  switch (images_tmp->state) {
  case SP_IMAGES_TMP_NOT_ALLOCATED:
    printf ("Structure 'SP_IMAGES_TMP' pas initialise\n");
    break;
  case SP_IMAGES_TMP_ALLOCATED:
    printf ("Structure 'SP_IMAGES_TMP' alloue :\n");

    /* tats de l'image des moyennes */
    switch (images_tmp->state_moyennes) {
    case SP_IMAGE_ALLOCATED:
      printf ("\tImage des moyennes simplement alloue\n");
      break;
    case SP_IMAGE_COMPUTED:
      printf ("\tImage des moyennes dtermine\n");
      break;
    default:
      return stereoError(ERR_SP_CAS_IMPOSSIBLE);
      /*     break; */
      /* Unreached */
    }

    /* tats de l'image des sigmas */
    switch (images_tmp->state_sigmas) {
    case SP_IMAGE_ALLOCATED:
      printf ("\tImage des sigmas simplement alloue\n");
      break;
    case SP_IMAGE_COMPUTED:
      printf ("\tImage des sigmas dtermine\n");
      break;
    default:
      return stereoError(ERR_SP_CAS_IMPOSSIBLE);
      /*     break; */
      /* Unreached */
    }
    break;

  default:
    return stereoError(ERR_SP_CAS_IMPOSSIBLE);
    /*     break; */
    /* Unreached */
  }

  return (STEREO_OK);
}

/***********************************************************************
 *
 * Fonction d'initialisation d'une structure 'SP_IMAGES_TMP'
 *
 ***********************************************************************/
int
init_images_tmp (SP_IMAGES_TMP*		images_tmp,
		 SP_IMAGES_SIZES*	images_sizes)

{
  /* Test de validit des paramtres */
  if (images_sizes->state != SP_IMAGES_SIZES_INITIALIZED)
    return stereoError(ERR_SP_IMAGES_SIZES_NOT_INITIALIZED);

  /* Copie de la structure 'images_sizes' dans le champ adquat de la
     structure images_tmp */
  memcpy (&(images_tmp->images_sizes), images_sizes, sizeof (SP_IMAGES_SIZES));

  /* Suivent toutes les allocations : les 2 images intermdiaires pour le
     filtrage... */
  images_tmp->intermediaire_haut
    = (short*) malloc (images_sizes->taille_finale * sizeof (short));
  if (images_tmp->intermediaire_haut == NULL)
    return stereoError(ERR_SP_MALLOC_IMAGES_TMP);
  images_tmp->state_intermediaire_haut = SP_IMAGE_ALLOCATED;
  images_tmp->intermediaire_bas
    = (short*) malloc (images_sizes->taille_finale * sizeof (short));
  if (images_tmp->intermediaire_bas == NULL)
    return stereoError(ERR_SP_MALLOC_IMAGES_TMP);
  images_tmp->state_intermediaire_bas = SP_IMAGE_ALLOCATED;

  /* ... les 2 images intermdiaires pour le cacul des moyennes et sigmas
     ... */
  images_tmp->intermediaire_moyennes1
    = (short*) malloc (images_sizes->taille_finale * sizeof (short));
  if (images_tmp->intermediaire_moyennes1 == NULL)
    return stereoError(ERR_SP_MALLOC_IMAGES_TMP);
  images_tmp->state_intermediaire_moyennes1 = SP_IMAGE_ALLOCATED;
  images_tmp->intermediaire_moyennes2
    = (int*) malloc (images_sizes->taille_finale * sizeof (int));
  if (images_tmp->intermediaire_moyennes2 == NULL)
    return stereoError(ERR_SP_MALLOC_IMAGES_TMP);
  images_tmp->state_intermediaire_moyennes2 = SP_IMAGE_ALLOCATED;

  /* ... les 2 images des moyennes et sigmas ... */
  images_tmp->im_moyennes
    = (float*) malloc (images_sizes->taille_finale * sizeof (float));
  if (images_tmp->im_moyennes == NULL)
    return stereoError(ERR_SP_MALLOC_IMAGES_TMP);
  images_tmp->state_moyennes = SP_IMAGE_ALLOCATED;
  images_tmp->im_sigmas
    = (float*) malloc (images_sizes->taille_finale * sizeof (float));
  if (images_tmp->im_sigmas == NULL)
    return stereoError(ERR_SP_MALLOC_IMAGES_TMP);
  images_tmp->state_sigmas = SP_IMAGE_ALLOCATED;


  images_tmp->state = SP_IMAGES_TMP_ALLOCATED;

  return (STEREO_OK);
}
